package fe

import (
	"fmt"
	"log"
	"math/rand"
	"net"
	"strings"
	"time"

	"git.autistici.org/ale/autoradio"
	"git.autistici.org/ale/autoradio/instrumentation"
	"github.com/miekg/dns"
)

var (
	// Max number of results for an A query.
	maxResults = 4

	// The names that we are serving. Currently, all services are
	// mapped to all the active nodes in the cluster.
	validNames = []string{
		"",
		"www",
		"stream",
		"etcd",
	}

	dnsQueryStats  = instrumentation.NewCounter("dns.status")
	dnsTargetStats = instrumentation.NewCounter("dns.target")
)

// DNS server.
type DnsRedirector struct {
	client         *autoradio.RadioAPI
	origin         string
	originNumParts int
	publicIp       string
	ttl            int
	soa            dns.RR
}

// NewDnsRedirector returns a DNS server for the given origin and
// publicIp. The A records served will have the specified ttl.
func NewDnsRedirector(client *autoradio.RadioAPI, origin, publicIp string, ttl int) *DnsRedirector {
	if !strings.HasSuffix(origin, ".") {
		origin += "."
	}

	// Create a SOA record for the zone. Some entries will be bogus.
	soa := &dns.SOA{
		Hdr: dns.RR_Header{
			Name:   origin,
			Rrtype: dns.TypeSOA,
			Class:  dns.ClassINET,
			Ttl:    43200,
		},
		Ns:      "ns." + origin,
		Mbox:    "hostmaster." + origin,
		Serial:  uint32(time.Now().Unix()),
		Refresh: 43200,
		Retry:   3600,
		Expire:  uint32(ttl),
		Minttl:  uint32(ttl),
	}

	return &DnsRedirector{
		client:         client,
		origin:         origin,
		originNumParts: len(dns.SplitDomainName(origin)),
		publicIp:       publicIp,
		ttl:            ttl,
		soa:            soa,
	}
}

// Randomly shuffle a list of strings.
func shuffle(list []string) []string {
	out := make([]string, len(list))
	for dst, src := range rand.Perm(len(list)) {
		out[dst] = list[src]
	}
	return out
}

func isValidQuery(query string) bool {
	for _, q := range validNames {
		if query == q {
			return true
		}
	}
	return false
}

// Create skeleton edns opt RR from the query and add it to the
// message m.
func ednsFromRequest(req, m *dns.Msg) {
	for _, r := range req.Extra {
		if r.Header().Rrtype == dns.TypeOPT {
			m.SetEdns0(4096, r.(*dns.OPT).Do())
			return
		}
	}
	return
}

// Create an A RR for a specific IP.
func (d *DnsRedirector) recordForIp(name string, ip string) *dns.A {
	var fqdn string
	if name == "" {
		fqdn = d.origin
	} else {
		fqdn = name + "." + d.origin
	}
	return &dns.A{
		Hdr: dns.RR_Header{
			Name:   fqdn,
			Rrtype: dns.TypeA,
			Class:  dns.ClassINET,
			Ttl:    uint32(d.ttl),
		},
		A: net.ParseIP(ip),
	}
}

// Strip the origin from the query.
func (d *DnsRedirector) getQuestionName(req *dns.Msg) string {
	lx := dns.SplitDomainName(req.Question[0].Name)
	ql := lx[0 : len(lx)-d.originNumParts]
	return strings.ToLower(strings.Join(ql, "."))
}

func (d *DnsRedirector) serveDNS(w dns.ResponseWriter, req *dns.Msg) {
	m := new(dns.Msg)

	// Just NACK ANYs
	if req.Question[0].Qtype == dns.TypeANY {
		m.SetRcode(req, dns.RcodeServerFailure)
		ednsFromRequest(req, m)
		w.WriteMsg(m)
		return
	}

	query := d.getQuestionName(req)
	var responseMsg string

	switch {
	case query == "" && req.Question[0].Qtype == dns.TypeSOA:
		// Serve SOA record.
		m.SetReply(req)
		m.MsgHdr.Authoritative = true
		m.Answer = append(m.Answer, d.soa)
		responseMsg = "SOA"

	case req.Question[0].Qtype == dns.TypeA:
		// Return an NXDOMAIN for unknown queries.
		if !isValidQuery(query) {
			m.SetRcode(req, dns.RcodeNameError)
			responseMsg = "NXDOMAIN"
			break
		}

		// Serve all active nodes on every request.
		ips, _ := d.client.GetNodeIPs()

		if ips == nil || len(ips) == 0 {
			// In case of errors retrieving the list of
			// active nodes, fall back to serving our
			// public IP (just to avoid returning an empty
			// reply, which might be cached for longer).
			ips = []string{d.publicIp}
		}

		// Shuffle the list in random order, and keep only the
		// first N results.
		ips = shuffle(ips)
		if len(ips) > maxResults {
			ips = ips[:maxResults]
		}

		m.SetReply(req)
		m.MsgHdr.Authoritative = true
		for _, ip := range ips {
			rec := d.recordForIp(query, ip)
			m.Answer = append(m.Answer, rec)
			dnsTargetStats.IncrVar(ipToMetric(ip))
		}
		responseMsg = fmt.Sprintf("%v", ips)

	default:
		// Return an error for anything else.
		m.SetRcode(req, dns.RcodeNameError)
		responseMsg = "NXDOMAIN"
	}

	if responseMsg == "NXDOMAIN" {
		dnsQueryStats.IncrVar("error")
	} else {
		dnsQueryStats.IncrVar("ok")
	}

	log.Printf("[%d] %s.%s %s (from %s) -> %s", req.MsgHdr.Id, query, d.origin, dns.TypeToString[req.Question[0].Qtype], w.RemoteAddr(), responseMsg)

	ednsFromRequest(req, m)
	w.WriteMsg(m)
}

// Run starts the DNS servers on the given address (both tcp and udp).
// It creates new goroutines and returns immediately.
func (d *DnsRedirector) Run(addr string) {
	dns.HandleFunc(d.origin, func(w dns.ResponseWriter, r *dns.Msg) {
		d.serveDNS(w, r)
	})

	for _, proto := range []string{"tcp", "udp"} {
		go func(proto string) {
			server := &dns.Server{Addr: addr, Net: proto}
			log.Printf("Starting DNS server on %s/%s", addr, proto)
			log.Fatal(server.ListenAndServe())
		}(proto)
	}
}

func ipToMetric(ip string) string {
	return strings.Replace(ip, ".", "_", -1)
}
