package node

import (
	"bytes"
	"encoding/xml"
	"io"
	"io/ioutil"
	"log"
	"net"
	"net/url"
	"os"
	"strconv"

	"git.autistici.org/ale/autoradio"
)

var (
	//shoutHttpPort = 8001
	maxClients = 10000

	icecastAdminPwFile = "/etc/icecast/.admin_pw"
	icecastAdminPw     string
)

// Return the global Icecast admin password. If it does not exist,
// generate a random one and save it for subsequent invocations.
func getIcecastAdminPassword() string {
	if icecastAdminPw == "" {
		data, err := ioutil.ReadFile(icecastAdminPwFile)
		if err != nil {
			icecastAdminPw = autoradio.GeneratePassword()
			if err := ioutil.WriteFile(icecastAdminPwFile, []byte(icecastAdminPw), 0600); err != nil {
				log.Printf("Couldn't save icecast admin password: %v", err)
			}
		} else {
			icecastAdminPw = string(data)
		}
	}
	return icecastAdminPw
}

type iceLimitsConfig struct {
	Clients int `xml:"clients"`
	Sources int `xml:"sources"`
	// Threadpool     int `xml:"threadpool"`
	QueueSize     int `xml:"queue-size"`
	ClientTimeout int `xml:"client-timeout"`
	HeaderTimeout int `xml:"header-timeout"`
	SourceTimeout int `xml:"source-timeout"`
	// BurstOnConnect int `xml:"burst-on-connect"`
	BurstSize int `xml:"burst-size"`
}

type iceAuthenticationConfig struct {
	SourcePassword string `xml:"source-password"`
	AdminUser      string `xml:"admin-user"`
	AdminPassword  string `xml:"admin-password"`
}

type iceListenConfig struct {
	BindAddress     string `xml:"bind-address"`
	Port            int    `xml:"port"`
	ShoutcastCompat int    `xml:"shoutcast-compat"`
}

type icePathsConfig struct {
	Basedir   string `xml:"basedir"`
	Logdir    string `xml:"logdir"`
	Webroot   string `xml:"webroot"`
	Adminroot string `xml:"adminroot"`
}

type iceLoggingConfig struct {
	Accesslog string `xml:"accesslog"`
	Errorlog  string `xml:"errorlog"`
	Loglevel  int    `xml:"loglevel"`
	Logsize   int    `xml:"logsize"`
}

type iceSecurityConfig struct {
	Chroot int `xml:"chroot"`
}

type iceRelayConfig struct {
	Server                 string `xml:"server"`
	Port                   int    `xml:"port"`
	Mount                  string `xml:"mount"`
	LocalMount             string `xml:"local-mount"`
	Username               string `xml:"username"`
	Password               string `xml:"password"`
	OnDemand               int    `xml:"on-demand"`
	RelayShoutcastMetadata int    `xml:"relay-shoutcast-metadata"`
}

type iceMountConfig struct {
	Name     string `xml:"mount-name"`
	Username string `xml:"username"`
	Password string `xml:"password"`
	// MaxListeners     int    `xml:"max-listeners"`
	FallbackMount    string `xml:"fallback-mount,omitempty"`
	FallbackOverride int    `xml:"fallback-override,omitempty"`
	Hidden           int    `xml:"hidden"`
	// NoYp             int    `xml:"no-yp"`
	OnConnect    string `xml:"on-connect,omitempty"`
	OnDisconnect string `xml:"on-disconnect,omitempty"`
}

// Configuration of the local Icecast daemon (meant for serialization
// to XML).
type icecastConfig struct {
	XMLName   xml.Name
	Limits    iceLimitsConfig         `xml:"limits"`
	Auth      iceAuthenticationConfig `xml:"authentication"`
	Hostname  string                  `xml:"hostname"`
	Fileserve int                     `xml:"fileserve"`
	Paths     icePathsConfig          `xml:"paths"`
	Logging   iceLoggingConfig        `xml:"logging"`
	Security  iceSecurityConfig       `xml:"security"`
	Listen    []iceListenConfig       `xml:"listen-socket"`
	Relays    []iceRelayConfig        `xml:"relay"`
	Mounts    []iceMountConfig        `xml:"mount"`
}

// Create an Icecast configuration suitable for a Debian-based system
// install of the 'icecast2' package. Things to note about the
// generated config:
//
// - It binds to the IcecastPort (defined in api.go) on all
// interfaces.
//
// - Local administration is practically disabled. A random admin
// password is created every time the daemon starts. Same goes for the
// global source password.
//
// Some of the parameters should probably be command-line flags, so
// that it is possible to set them on a per-host basis.
//
func defaultDebianConfig(publicIp string) *icecastConfig {
	// Set the icecast admin password once, on the first run, and
	// save it on the filesystem. We don't use the global source
	// password, but icecast is happier if it's set, so we just
	// use a random password every time.
	sourcePw := autoradio.GeneratePassword()
	adminPw := getIcecastAdminPassword()

	return &icecastConfig{
		XMLName: xml.Name{"", "icecast"},
		Limits: iceLimitsConfig{
			Clients: maxClients,
			Sources: maxClients / 2,
			// Threadpool:     16,
			QueueSize:     1 << 20,
			ClientTimeout: 30,
			HeaderTimeout: 15,
			SourceTimeout: 60,
			// BurstOnConnect: 1,
			BurstSize: 131072,
		},
		Auth: iceAuthenticationConfig{
			SourcePassword: sourcePw,
			AdminUser:      "admin",
			AdminPassword:  adminPw,
		},
		Hostname:  publicIp,
		Fileserve: 1,
		Paths: icePathsConfig{
			Basedir:   "/usr/share/icecast2",
			Logdir:    "/var/log/icecast2",
			Webroot:   "/usr/share/icecast2/web",
			Adminroot: "/usr/share/icecast2/admin",
		},
		Logging: iceLoggingConfig{
			Accesslog: "access.log",
			Errorlog:  "error.log",
			Loglevel:  3,
			Logsize:   10000,
		},
		Security: iceSecurityConfig{0},
		Listen: []iceListenConfig{
			{"0.0.0.0", autoradio.IcecastPort, 0},
			//{"0.0.0.0", shoutHttpPort, 1},
		},
	}
}

func newIcecastConfig(publicIp string) *icecastConfig {
	return defaultDebianConfig(publicIp)
}

// Encode the configuration to XML.
func (c *icecastConfig) Encode() ([]byte, error) {
	var buf bytes.Buffer

	output, err := xml.MarshalIndent(c, "", "  ")
	if err != nil {
		return nil, err
	}
	io.WriteString(&buf, "<!-- Automatically generated, do not edit -->\n\n")
	if _, err := buf.Write(output); err != nil {
		return nil, err
	}
	return buf.Bytes(), nil
}

// EncodeToFile writes the configuration to a file.
func (c *icecastConfig) EncodeToFile(path string) error {
	file, err := os.Create(path)
	if err != nil {
		return err
	}
	defer file.Close()

	data, err := c.Encode()
	if err != nil {
		return err
	}
	_, err = file.Write(data)
	return err
}

func mountToConfig(m *autoradio.Mount) iceMountConfig {
	mconfig := iceMountConfig{
		Name:     autoradio.MountNameToIcecastPath(m.Name),
		Username: m.Username,
		Password: m.Password,
		Hidden:   0,
		// MaxListeners: 1000,
		// NoYp:   1,
	}
	if m.Fallback != "" {
		mconfig.FallbackMount = m.Fallback
		mconfig.FallbackOverride = 1
	}
	return mconfig
}

func relayToConfig(m *autoradio.Mount) (iceRelayConfig, bool) {
	u, err := url.Parse(m.RelayUrl)
	if err != nil {
		// A failure here is almost invisible and not very
		// useful, but at least we can prevent garbling the
		// resulting icecast config.
		log.Printf("can't parse relay url for %s: %s", m.Name, err)
		return iceRelayConfig{}, false
	}
	server, port, err := net.SplitHostPort(u.Host)
	if err != nil {
		server = u.Host
		port = "80"
	}
	iport, _ := strconv.Atoi(port)

	rc := iceRelayConfig{
		Mount:                  u.Path,
		LocalMount:             autoradio.MountNameToIcecastPath(m.Name),
		Server:                 server,
		Port:                   iport,
		OnDemand:               1,
		RelayShoutcastMetadata: 1,
	}
	if u.User != nil {
		rc.Username = u.User.Username()
		if p, ok := u.User.Password(); ok {
			rc.Password = p
		}
	}
	return rc, true
}

func mountToRelayConfig(masterAddr string, m *autoradio.Mount) iceRelayConfig {
	return iceRelayConfig{
		Mount:                  autoradio.MountNameToIcecastPath(m.Name),
		LocalMount:             autoradio.MountNameToIcecastPath(m.Name),
		Server:                 masterAddr,
		Port:                   autoradio.IcecastPort,
		Username:               m.Username,
		Password:               m.Password,
		OnDemand:               1,
		RelayShoutcastMetadata: 1,
	}
}

// Update the configuration with the current list of mounts and
// masterelection state. This will clear the Mounts and Relays fields
// and set them to new values.
func (ic *icecastConfig) Update(config *ClusterConfig, isMaster bool, masterAddr string) {
	ic.Mounts = nil
	ic.Relays = nil
	mounts := make([]iceMountConfig, 0)
	relays := make([]iceRelayConfig, 0)

	for _, m := range config.ListMounts() {
		switch {
		case m.IsRelay():
			if rc, ok := relayToConfig(m); ok {
				relays = append(relays, rc)
			}
		case isMaster:
			mounts = append(mounts, mountToConfig(m))
		default:
			relays = append(relays, mountToRelayConfig(masterAddr, m))
		}
	}

	if len(mounts) > 0 {
		ic.Mounts = mounts
	}
	if len(relays) > 0 {
		ic.Relays = relays
	}
}
